#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # 1A.1 - Dictionnaires, fonctions, code de Vigenère (correction)
# 
# Le notebook ne fait que crypter et décrypter un message sachant le code connu. Casser le code requiert quelques astuces décrites dnas ce notebook : [casser le code de Vigenère](http://www.xavierdupre.fr/app/ensae_teaching_cs/helpsphinx/notebooks/expose_vigenere.html).

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# ### Exercice 1

# In[2]:


def lettre_suivante(lettre) :
    c = ord(lettre) - ord('a')
    c = (c + 1) % 26
    return chr (c + ord('a'))
    
print (lettre_suivante('m'), lettre_suivante('z'))


# ### Exercice 2

# In[3]:


mots = ['eddard', 'catelyn', 'robb', 'sansa', 'arya', 'brandon',
        'rickon', 'theon', 'rorbert', 'cersei', 'tywin', 'jaime',
        'tyrion', 'shae', 'bronn', 'lancel', 'joffrey', 'sandor',
        'varys', 'renly', 'a' ]
        
def mots_lettre_position (liste, lettre, position) :
    res = [ ]
    for mot in liste :
        if position < len(mot) and mot[position] == lettre :
            res.append (mot)
    return res
    
r = mots_lettre_position ( mots, 'y', 1)
print (r)


# ### Exercice 3 : utilisation d'un dictionnaire

# L'énoncé suggère d'utiliser comme clé de dictionnaire le couple ``(position, lettre)`` et la fonction doit retourne la liste des mots qui ont tous la même lettre à la même position. Le dictionnaire ``dictionnaire_bien_choisi`` de l'énoncé doit avoir pour clés des couples ``(position, lettre)`` et pour valeurs des listes de prénoms.

# In[4]:


def dictionnaire_choisi (liste) :
    d = { }
    for mot in liste :
        for i,c in enumerate(mot) :
            d [i,c] = d.get ((i,c), []) + [ mot ]
    return d
    
def mots_lettre_position (d, lettre, position) :
    return d.get ( (position, lettre), [] )

d = dictionnaire_choisi(mots)
r = mots_lettre_position ( d, 'y', 1)
print ("résultat=",r)
print ("dictionnaire=",d)


# S'il permet d'aller beaucoup plus vite pour effectuer une recherche, le dictionnaire ``d`` contient beaucoup plus de mots que la liste initiale. Si on suppose que tous les mots sont uniques, il en contient exactement autant que la somme des longueurs de chaque mot.
# 
# **A quoi ça sert ?** Tout dépend du nombre de fois qu'on n'effectue ce type de **recherche**. Il faut d'abord décomposer les deux méthodes en coût fixe (préparation du dictionnaire) et coût recherche puis regarder la page [Time Complexity](https://wiki.python.org/moin/TimeComplexity). On obtient :
# 
# - liste de l'exercice 2 :  coût fixe = 0, coût variable $\sim O(N)$ 
# - dictionaire de l'exercice 3 : coût fixe $\sim O(L)$, coût variable $\sim O(1)$ 
# 
# Où :
# 
# - $N$ est le nombre de mots,
# - $L$ est la somme des nombres de lettres de chaque mot,
# - $M$ est la longueur maximale d'un mot.
# 
# Les dictionnaires en Python utilisent une [table de hashage](http://fr.wikipedia.org/wiki/Table_de_hachage) pour stocker les clés. L'objet ``map`` de Python ne rapproche plus de l'objet ``unordered_map`` de C++ que de l'objet ``map``. Ce dernier (C++ uniquement) est un tableau trié. L'accès à chaque élément se fait par dichotomie en $O(\ln_2 n)$ (voir [Standard C++ Containers](http://www.cs.northwestern.edu/~riesbeck/programming/c++/stl-summary.html#map). Le coût dans ce cas serait (toujours en C++) :
# 
# - dictionaire de l'exercice 3 : coût fixe $\sim O(L \, ln_2(26 * M))$, coût variable $\sim O(ln_2(26 * M))$ 
# 
# 
# Si on effectue cette recherche un grand nombre de fois, l'utilisation d'un dictionnaire permet d'être beaucoup plus rapide même si on doit créer une structure intermédiaire. Ce schéma revient régulièrement : **représenter autrement les données pour accélérer un traitement effectué un grand nombre de fois**. 
# 
# Vous pouvez lire également :
# 
# - [hash](https://docs.python.org/3.4/reference/datamodel.html#object.__hash__)
# - [STL Container Performance ](http://john-ahlgren.blogspot.fr/2013/10/stl-container-performance.html)
# - [C++11: unordered_map vs map](http://kariddi.blogspot.fr/2012/07/c11-unorderedmap-vs-map.html)
# - [AVL tree](http://en.wikipedia.org/wiki/AVL_tree)
# - [List of data structures](http://en.wikipedia.org/wiki/List_of_data_structures)
# - [Time complexity of accessing a Python dict](http://stackoverflow.com/questions/1963507/time-complexity-of-accessing-a-python-dict)
# - [Hash Table Performance Tests](http://preshing.com/20110603/hash-table-performance-tests/)
# - [How to implement a good __hash__ function in python](http://stackoverflow.com/questions/4005318/how-to-implement-a-good-hash-function-in-python)

# ### Exercice 4 : crypter et décrypter selon Vigenère
# 
# Tout d'abord le code de César :

# In[5]:


def code_cesar(m):
    s = "".join( [ chr((ord(l)-65+3)%26+65) for l in m ] )
    return s

m = "JENESUISPASCODE"
print(code_cesar(m))


# Et le code de Vigenère :

# In[6]:


def code_vigenere ( message, cle) :
    message_code = ""
    for i,c in enumerate(message) :
        d = cle[ i % len(cle) ]
        d = ord(d) - 65
        message_code += chr((ord(c)-65+d)%26+65)
    return message_code

m = "JENESUISPASCODE"
print ( code_vigenere (m, "DOP") )


# Et le décryptage du code de Vigenère pour lequel on modifie la fonction précédente qui pourra alors coder et décoder.

# In[7]:


def code_vigenere ( message, cle, decode = False) :      # ligne changée
    message_code = ""
    for i,c in enumerate(message) :
        d = cle[ i % len(cle) ]
        d = ord(d) - 65
        if decode : d = 26 - d                          # ligne ajoutée
        message_code += chr((ord(c)-65+d)%26+65)
    return message_code
    
m = "JENESUISPASCODE"
c = code_vigenere (m, "DOP") 
d = code_vigenere (c, "DOP", True) 
print(c,d)


# Pour retrouver le code de César, il suffit de choisir une clé d'une seule lettre :

# In[8]:


c = code_vigenere (m, "D") 
print(c)


# On peut casser le code de Vigenère. Vous trouverez la solution ici : [casser le code de Vigenère](http://www.xavierdupre.fr/app/ensae_teaching_cs/helpsphinx/notebooks/expose_vigenere.html).

# In[9]: